@@ -27,6 +27,16 @@ def SpecialWindow(cmd, stdin=None, override_title=None, cwd_from=None, cwd=None,
     return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env)
 
 
+def add_active_id_to_history(items, item_id, maxlen=64):
+    try:
+        items.remove(item_id)
+    except ValueError:
+        pass
+    items.append(item_id)
+    if len(items) > maxlen:
+        items.popleft()
+
+
 class Tab:  # {{{
 
     def __init__(self, tab_manager, session_tab=None, special_window=None, cwd_from=None):
@@ -34,21 +44,24 @@ class Tab:  # {{{
         self.tab_manager_ref = weakref.ref(tab_manager)
         self.os_window_id = tab_manager.os_window_id
         self.id = add_tab(self.os_window_id)
+        self.active_window_history = deque()
         if not self.id:
             raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
         self.opts, self.args = tab_manager.opts, tab_manager.args
-        self.margin_width, self.padding_width = pt_to_px(
-                self.opts.window_margin_width, self.os_window_id), pt_to_px(self.opts.window_padding_width, self.os_window_id)
+        self.margin_width, self.padding_width, self.single_window_margin_width = map(
+            lambda x: pt_to_px(getattr(self.opts, x), self.os_window_id), (
+                'window_margin_width', 'window_padding_width', 'single_window_margin_width'))
         self.name = getattr(session_tab, 'name', '')
         self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or self.opts.enabled_layouts]
         self.borders = Borders(self.os_window_id, self.id, self.opts, pt_to_px(self.opts.window_border_width, self.os_window_id), self.padding_width)
         self.windows = deque()
         for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
             setattr(self, which + '_window', partial(self.nth_window, num=i))
+        self._last_used_layout = self._current_layout_name = None
         if session_tab is None:
             self.cwd = self.args.directory
             sl = self.enabled_layouts[0]
-            self.current_layout = self.create_layout_object(sl)
+            self._set_current_layout(sl)
             if special_window is None:
                 self.new_window(cwd_from=cwd_from)
             else:
@@ -56,9 +69,14 @@ class Tab:  # {{{
         else:
             self.cwd = session_tab.cwd or self.args.directory
             l0 = session_tab.layout
-            self.current_layout = self.create_layout_object(l0)
+            self._set_current_layout(l0)
             self.startup(session_tab)
 
+    def _set_current_layout(self, layout_name):
+        self._last_used_layout = self._current_layout_name
+        self.current_layout = self.create_layout_object(layout_name)
+        self._current_layout_name = layout_name
+
     def startup(self, session_tab):
         for cmd in session_tab.windows:
             if isinstance(cmd, (SpecialWindowInstance,)):
@@ -77,6 +95,9 @@ class Tab:  # {{{
             old_active_window = self.windows[self._active_window_idx]
         except Exception:
             old_active_window = None
+        else:
+            wid = old_active_window.id if old_active_window.overlay_for is None else old_active_window.overlay_for
+            add_active_id_to_history(self.active_window_history, wid)
         self._active_window_idx = max(0, min(val, len(self.windows) - 1))
         try:
             new_active_window = self.windows[self._active_window_idx]
@@ -139,7 +160,10 @@ class Tab:  # {{{
                 w.change_titlebar_color()
 
     def create_layout_object(self, name):
-        return create_layout_object_for(name, self.os_window_id, self.id, self.margin_width, self.padding_width, self.borders.border_width)
+        return create_layout_object_for(
+            name, self.os_window_id, self.id, self.margin_width,
+            self.single_window_margin_width, self.padding_width,
+            self.borders.border_width)
 
     def next_layout(self):
         if len(self.enabled_layouts) > 1:
@@ -150,15 +174,22 @@ class Tab:  # {{{
             else:
                 idx = -1
             nl = self.enabled_layouts[(idx + 1) % len(self.enabled_layouts)]
-            self.current_layout = self.create_layout_object(nl)
+            self._set_current_layout(nl)
             self.relayout()
 
-    def goto_layout(self, layout_name):
+    def last_used_layout(self):
+        if len(self.enabled_layouts) > 1 and self._last_used_layout and self._last_used_layout != self._current_layout_name:
+            self._set_current_layout(self._last_used_layout)
+            self.relayout()
+
+    def goto_layout(self, layout_name, raise_exception=False):
         layout_name = layout_name.lower()
         if layout_name not in self.enabled_layouts:
+            if raise_exception:
+                raise ValueError(layout_name)
             log_error('Unknown or disabled layout: {}'.format(layout_name))
             return
-        self.current_layout = self.create_layout_object(layout_name)
+        self._set_current_layout(layout_name)
         self.relayout()
 
     def resize_window_by(self, window_id, increment, is_horizontal):
@@ -238,7 +269,19 @@ class Tab:  # {{{
 
     def nth_window(self, num=0):
         if self.windows:
-            self.active_window_idx = self.current_layout.nth_window(self.windows, num)
+            if num < 0:
+                try:
+                    old_window_id = self.active_window_history[num]
+                except IndexError:
+                    return
+                for idx, w in enumerate(self.windows):
+                    if w.id == old_window_id:
+                        self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
+                        break
+                else:
+                    return
+            else:
+                self.active_window_idx = self.current_layout.nth_window(self.windows, num)
             self.relayout_borders()
             glfw_post_empty_event()
 
@@ -254,6 +297,16 @@ class Tab:  # {{{
     def previous_window(self):
         self._next_window(-1)
 
+    prev_window = previous_window
+
+    def neighboring_window(self, which):
+        neighbors = self.current_layout.neighbors(self.windows, self.active_window_idx)
+        candidates = neighbors.get(which)
+        if candidates:
+            self.active_window_idx = self.current_layout.set_active_window(self.windows, candidates[0])
+            self.relayout_borders()
+            glfw_post_empty_event()
+
     def move_window(self, delta=1):
         self.active_window_idx = self.current_layout.move_window(self.windows, self.active_window_idx, delta)
         self.relayout()
@@ -268,9 +321,9 @@ class Tab:  # {{{
     def move_window_backward(self):
         self.move_window(-1)
 
-    def list_windows(self):
+    def list_windows(self, active_window):
         for w in self:
-            yield w.as_dict()
+            yield w.as_dict(is_focused=w is active_window)
 
     def matches(self, field, pat):
         if field == 'id':
@@ -303,7 +356,9 @@ class TabManager:  # {{{
 
     def __init__(self, os_window_id, opts, args, startup_session):
         self.os_window_id = os_window_id
+        self.last_active_tab_id = None
         self.opts, self.args = opts, args
+        self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
         self.tabs = []
         self.active_tab_history = deque()
         self.tab_bar = TabBar(self.os_window_id, opts)
@@ -321,11 +376,10 @@ class TabManager:  # {{{
     def active_tab_idx(self, val):
         try:
             old_active_tab = self.tabs[self._active_tab_idx]
-            self.active_tab_history.append(old_active_tab.id)
-            if len(self.active_tab_history) > 64:
-                self.active_tab_history.popleft()
         except Exception:
             old_active_tab = None
+        else:
+            add_active_id_to_history(self.active_tab_history, old_active_tab.id)
         self._active_tab_idx = max(0, min(val, len(self.tabs) - 1))
         try:
             new_active_tab = self.tabs[self._active_tab_idx]
@@ -342,7 +396,8 @@ class TabManager:  # {{{
                     w.focus_changed(True)
 
     def refresh_sprite_positions(self):
-        self.tab_bar.screen.refresh_sprite_positions()
+        if not self.tab_bar_hidden:
+            self.tab_bar.screen.refresh_sprite_positions()
 
     def _add_tab(self, tab):
         before = len(self.tabs)
@@ -362,12 +417,13 @@ class TabManager:  # {{{
         set_active_tab(self.os_window_id, idx)
 
     def tabbar_visibility_changed(self):
-        self.tab_bar.layout()
-        self.resize(only_tabs=True)
-        glfw_post_empty_event()
+        if not self.tab_bar_hidden:
+            self.tab_bar.layout()
+            self.resize(only_tabs=True)
+            glfw_post_empty_event()
 
     def mark_tab_bar_dirty(self):
-        if len(self.tabs) > 1:
+        if len(self.tabs) > 1 and not self.tab_bar_hidden:
             mark_tab_bar_dirty(self.os_window_id)
 
     def update_tab_bar_data(self):
@@ -375,8 +431,9 @@ class TabManager:  # {{{
 
     def resize(self, only_tabs=False):
         if not only_tabs:
-            self.tab_bar.layout()
-            self.mark_tab_bar_dirty()
+            if not self.tab_bar_hidden:
+                self.tab_bar.layout()
+                self.mark_tab_bar_dirty()
         for tab in self.tabs:
             tab.relayout()
 
@@ -399,6 +456,15 @@ class TabManager:  # {{{
     def goto_tab(self, tab_num):
         if tab_num < len(self.tabs) and 0 <= tab_num:
             self.set_active_tab_idx(tab_num)
+        elif tab_num < 0:
+            try:
+                old_active_tab_id = self.active_tab_history[tab_num]
+            except IndexError:
+                return
+            for idx, tab in enumerate(self.tabs):
+                if tab.id == old_active_tab_id:
+                    self.set_active_tab_idx(idx)
+                    break
 
     def __iter__(self):
         return iter(self.tabs)
@@ -406,12 +472,14 @@ class TabManager:  # {{{
     def __len__(self):
         return len(self.tabs)
 
-    def list_tabs(self):
+    def list_tabs(self, active_tab, active_window):
         for tab in self:
             yield {
                 'id': tab.id,
+                'is_focused': tab is active_tab,
                 'title': tab.name or tab.title,
-                'windows': list(tab.list_windows()),
+                'layout': tab.current_layout.name,
+                'windows': list(tab.list_windows(active_window)),
             }
 
     @property
@@ -438,10 +506,16 @@ class TabManager:  # {{{
             self._set_active_tab(nidx)
             self.mark_tab_bar_dirty()
 
-    def new_tab(self, special_window=None, cwd_from=None):
+    def new_tab(self, special_window=None, cwd_from=None, as_neighbor=False):
+        nidx = self.active_tab_idx + 1
         idx = len(self.tabs)
         self._add_tab(Tab(self, special_window=special_window, cwd_from=cwd_from))
         self._set_active_tab(idx)
+        if len(self.tabs) > 2 and as_neighbor and idx != nidx:
+            self.tabs[idx], self.tabs[nidx] = self.tabs[nidx], self.tabs[idx]
+            swap_tabs(self.os_window_id, idx, nidx)
+            self._set_active_tab(nidx)
+            idx = nidx
         self.mark_tab_bar_dirty()
         return self.tabs[idx]
 
