@@ -15,16 +15,13 @@ from time import monotonic
 from .constants import (
     appname, is_macos, is_wayland, supports_primary_selection
 )
-from .fast_data_types import (
-    GLSL_VERSION, close_tty, log_error_string, open_tty, redirect_std_streams,
-    x11_display
-)
 from .rgb import Color, to_color
 
 BASE = os.path.dirname(os.path.abspath(__file__))
 
 
 def load_shaders(name):
+    from .fast_data_types import GLSL_VERSION
     vert = open(os.path.join(BASE, '{}_vertex.glsl'.format(name))).read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
     frag = open(os.path.join(BASE, '{}_fragment.glsl'.format(name))).read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
     return vert, frag
@@ -38,8 +35,9 @@ def safe_print(*a, **k):
 
 
 def log_error(*a, **k):
+    from .fast_data_types import log_error_string
     try:
-        msg = k.get('sep', ' ').join(map(str, a)) + k.get('end', '\n')
+        msg = k.get('sep', ' ').join(map(str, a)) + k.get('end', '')
         log_error_string(msg.replace('\0', ''))
     except Exception:
         pass
@@ -63,6 +61,9 @@ def color_from_int(val):
 
 def parse_color_set(raw):
     parts = raw.split(';')
+    lp = len(parts)
+    if lp % 2 != 0:
+        return
     for c, spec in [parts[i:i + 2] for i in range(0, len(parts), 2)]:
         try:
             c = int(c)
@@ -180,6 +181,7 @@ def detach(fork=True, setsid=True, redirect=True):
     if setsid:
         os.setsid()
     if redirect:
+        from .fast_data_types import redirect_std_streams
         redirect_std_streams(os.devnull)
 
 
@@ -192,9 +194,10 @@ def adjust_line_height(cell_height, val):
 def init_startup_notification_x11(window_handle, startup_id=None):
     # https://specifications.freedesktop.org/startup-notification-spec/startup-notification-latest.txt
     from kitty.fast_data_types import init_x11_startup_notification
-    sid = startup_id or os.environ.pop('DESKTOP_STARTUP_ID', None)  # ensure child processes dont get this env var
+    sid = startup_id or os.environ.pop('DESKTOP_STARTUP_ID', None)  # ensure child processes don't get this env var
     if not sid:
         return
+    from .fast_data_types import x11_display
     display = x11_display()
     if not display:
         return
@@ -369,10 +372,12 @@ def write_all(fd, data):
 class TTYIO:
 
     def __enter__(self):
-        self.tty_fd, self.original_termios = open_tty()
+        from .fast_data_types import open_tty
+        self.tty_fd, self.original_termios = open_tty(True)
         return self
 
     def __exit__(self, *a):
+        from .fast_data_types import close_tty
         close_tty(self.tty_fd, self.original_termios)
 
     def send(self, data):
@@ -382,22 +387,48 @@ class TTYIO:
             for chunk in data:
                 write_all(self.tty_fd, chunk)
 
-    def recv(self, more_needed, timeout):
-        from io import DEFAULT_BUFFER_SIZE
-        import select
+    def recv(self, more_needed, timeout, sz=1):
         fd = self.tty_fd
         start_time = monotonic()
         while timeout > monotonic() - start_time:
-            rd = select.select([fd], [], [], max(0, timeout - (monotonic() - start_time)))[0]
-            if not rd:
-                break
-            data = os.read(fd, DEFAULT_BUFFER_SIZE)
-            if not data:
-                break  # eof
-            if not more_needed(data):
+            # will block for 0.1 secs waiting for data because we have set
+            # VMIN=0 VTIME=1 in termios
+            data = os.read(fd, sz)
+            if data and not more_needed(data):
                 break
 
 
+def natsort_ints(iterable):
+
+    def convert(text):
+        return int(text) if text.isdigit() else text
+
+    def alphanum_key(key):
+        return tuple(map(convert, re.split(r'(\d+)', key)))
+
+    return sorted(iterable, key=alphanum_key)
+
+
+def exe_exists(exe):
+    for loc in os.environ.get('PATH', '').split(os.pathsep):
+        if loc and os.access(os.path.join(loc, exe), os.X_OK):
+            return os.path.join(loc, exe)
+    return False
+
+
 def get_editor():
-    import shlex
-    return shlex.split(os.environ.get('EDITOR', 'vim'))
+    ans = getattr(get_editor, 'ans', False)
+    if ans is False:
+        import shlex
+        ans = os.environ.get('EDITOR')
+        if not ans or not exe_exists(shlex.split(ans)[0]):
+            for q in ('vim', 'nvim', 'vi', 'emacs', 'micro', 'nano', 'vis'):
+                r = exe_exists(q)
+                if r:
+                    ans = r
+                    break
+            else:
+                ans = 'vim'
+        ans = shlex.split(ans)
+        get_editor.ans = ans
+    return ans
